@@ -20,8 +20,8 @@ describe :kernel_require_basic, shared: true do
 
     it "raises a LoadError if the file does not exist" do
       path = File.expand_path "nonexistent.rb", CODE_LOADING_DIR
-      File.exist?(path).should be_false
-      lambda { @object.send(@method, path) }.should raise_error(LoadError)
+      File.should_not.exist?(path)
+      -> { @object.send(@method, path) }.should raise_error(LoadError)
       ScratchPad.recorded.should == []
     end
 
@@ -41,8 +41,8 @@ describe :kernel_require_basic, shared: true do
           end
 
           it "raises a LoadError" do
-            File.exist?(@path).should be_true
-            lambda { @object.send(@method, @path) }.should raise_error(LoadError)
+            File.should.exist?(@path)
+            -> { @object.send(@method, @path) }.should raise_error(LoadError)
           end
         end
       end
@@ -57,19 +57,19 @@ describe :kernel_require_basic, shared: true do
     end
 
     it "raises a TypeError if passed nil" do
-      lambda { @object.send(@method, nil) }.should raise_error(TypeError)
+      -> { @object.send(@method, nil) }.should raise_error(TypeError)
     end
 
     it "raises a TypeError if passed a Fixnum" do
-      lambda { @object.send(@method, 42) }.should raise_error(TypeError)
+      -> { @object.send(@method, 42) }.should raise_error(TypeError)
     end
 
     it "raises a TypeError if passed an Array" do
-      lambda { @object.send(@method, []) }.should raise_error(TypeError)
+      -> { @object.send(@method, []) }.should raise_error(TypeError)
     end
 
     it "raises a TypeError if passed an object that does not provide #to_str" do
-      lambda { @object.send(@method, mock("not a filename")) }.should raise_error(TypeError)
+      -> { @object.send(@method, mock("not a filename")) }.should raise_error(TypeError)
     end
 
     it "raises a TypeError if passed an object that has #to_s but not #to_str" do
@@ -77,14 +77,14 @@ describe :kernel_require_basic, shared: true do
       name.stub!(:to_s).and_return("load_fixture.rb")
       $LOAD_PATH << "."
       Dir.chdir CODE_LOADING_DIR do
-        lambda { @object.send(@method, name) }.should raise_error(TypeError)
+        -> { @object.send(@method, name) }.should raise_error(TypeError)
       end
     end
 
     it "raises a TypeError if #to_str does not return a String" do
       name = mock("#to_str returns nil")
       name.should_receive(:to_str).at_least(1).times.and_return(nil)
-      lambda { @object.send(@method, name) }.should raise_error(TypeError)
+      -> { @object.send(@method, name) }.should raise_error(TypeError)
     end
 
     it "calls #to_path on non-String objects" do
@@ -163,14 +163,14 @@ describe :kernel_require_basic, shared: true do
     it "does not require file twice after $LOAD_PATH change" do
       $LOAD_PATH << CODE_LOADING_DIR
       @object.require("load_fixture.rb").should be_true
-      $LOAD_PATH.unshift CODE_LOADING_DIR + "/gem"
+      $LOAD_PATH.push CODE_LOADING_DIR + "/gem"
       @object.require("load_fixture.rb").should be_false
       ScratchPad.recorded.should == [:loaded]
     end
 
     it "does not resolve a ./ relative path against $LOAD_PATH entries" do
       $LOAD_PATH << CODE_LOADING_DIR
-      lambda do
+      -> do
         @object.send(@method, "./load_fixture.rb")
       end.should raise_error(LoadError)
       ScratchPad.recorded.should == []
@@ -178,7 +178,7 @@ describe :kernel_require_basic, shared: true do
 
     it "does not resolve a ../ relative path against $LOAD_PATH entries" do
       $LOAD_PATH << CODE_LOADING_DIR
-      lambda do
+      -> do
         @object.send(@method, "../code/load_fixture.rb")
       end.should raise_error(LoadError)
       ScratchPad.recorded.should == []
@@ -208,14 +208,14 @@ describe :kernel_require, shared: true do
     # intentional for security reasons.
     it "does not load a bare filename unless the current working directory is in $LOAD_PATH" do
       Dir.chdir CODE_LOADING_DIR do
-        lambda { @object.require("load_fixture.rb") }.should raise_error(LoadError)
+        -> { @object.require("load_fixture.rb") }.should raise_error(LoadError)
         ScratchPad.recorded.should == []
       end
     end
 
     it "does not load a relative path unless the current working directory is in $LOAD_PATH" do
       Dir.chdir File.dirname(CODE_LOADING_DIR) do
-        lambda do
+        -> do
           @object.require("code/load_fixture.rb")
         end.should raise_error(LoadError)
         ScratchPad.recorded.should == []
@@ -242,12 +242,21 @@ describe :kernel_require, shared: true do
       @object.require("load_fixture").should be_true
       ScratchPad.recorded.should == [:loaded]
     end
+
+    ruby_bug "#16926", "2.7"..."3.0" do
+      it "does not load a feature twice when $LOAD_PATH has been modified" do
+        $LOAD_PATH.replace [CODE_LOADING_DIR]
+        @object.require("load_fixture").should be_true
+        $LOAD_PATH.replace [File.expand_path("b", CODE_LOADING_DIR), CODE_LOADING_DIR]
+        @object.require("load_fixture").should be_false
+      end
+    end
   end
 
   describe "(file extensions)" do
     it "loads a .rb extensioned file when passed a non-extensioned path" do
       path = File.expand_path "load_fixture", CODE_LOADING_DIR
-      File.exist?(path).should be_true
+      File.should.exist?(path)
       @object.require(path).should be_true
       ScratchPad.recorded.should == [:loaded]
     end
@@ -271,7 +280,7 @@ describe :kernel_require, shared: true do
 
     it "loads a .rb extensioned file when passed a non-.rb extensioned path" do
       path = File.expand_path "load_fixture.ext", CODE_LOADING_DIR
-      File.exist?(path).should be_true
+      File.should.exist?(path)
       @object.require(path).should be_true
       ScratchPad.recorded.should == [:loaded]
     end
@@ -354,26 +363,13 @@ describe :kernel_require, shared: true do
           rm_r @dir, @symlink_to_dir
         end
 
-        ruby_version_is ""..."2.4.4" do
-          it "canonicalizes neither the entry in $LOAD_PATH nor the filename passed to #require" do
-            $LOAD_PATH.unshift(@symlink_to_dir)
-            @object.require("symfile").should be_true
-            loaded_feature = "#{@symlink_to_dir}/symfile.rb"
-            ScratchPad.recorded.should == [loaded_feature]
-            $".last.should == loaded_feature
-            $LOAD_PATH[0].should == @symlink_to_dir
-          end
-        end
-
-        ruby_version_is "2.4.4" do
-          it "canonicalizes the entry in $LOAD_PATH but not the filename passed to #require" do
-            $LOAD_PATH.unshift(@symlink_to_dir)
-            @object.require("symfile").should be_true
-            loaded_feature = "#{@dir}/symfile.rb"
-            ScratchPad.recorded.should == [loaded_feature]
-            $".last.should == loaded_feature
-            $LOAD_PATH[0].should == @symlink_to_dir
-          end
+        it "canonicalizes the entry in $LOAD_PATH but not the filename passed to #require" do
+          $LOAD_PATH.unshift(@symlink_to_dir)
+          @object.require("symfile").should be_true
+          loaded_feature = "#{@dir}/symfile.rb"
+          ScratchPad.recorded.should == [loaded_feature]
+          $".last.should == loaded_feature
+          $LOAD_PATH[0].should == @symlink_to_dir
         end
       end
     end
@@ -381,7 +377,7 @@ describe :kernel_require, shared: true do
     it "does not store the path if the load fails" do
       $LOAD_PATH << CODE_LOADING_DIR
       saved_loaded_features = $LOADED_FEATURES.dup
-      lambda { @object.require("raise_fixture.rb") }.should raise_error(RuntimeError)
+      -> { @object.require("raise_fixture.rb") }.should raise_error(RuntimeError)
       $LOADED_FEATURES.should == saved_loaded_features
     end
 
@@ -527,41 +523,25 @@ describe :kernel_require, shared: true do
       ScratchPad.recorded.should == []
     end
 
-    ruby_version_is ""..."2.5" do
-      it "complex, enumerator, rational, thread and unicode_normalize are already required" do
-        provided = %w[complex enumerator rational thread unicode_normalize]
-        features = ruby_exe("puts $LOADED_FEATURES", options: '--disable-gems')
-        provided.each { |feature|
-          features.should =~ /\b#{feature}\.(rb|so|jar)$/
-        }
+    it "complex, enumerator, rational and thread are already required" do
+      provided = %w[complex enumerator rational thread]
+      features = ruby_exe("puts $LOADED_FEATURES", options: '--disable-gems')
+      provided.each { |feature|
+        features.should =~ /\b#{feature}\.(rb|so|jar)$/
+      }
 
-        code = provided.map { |f| "puts require #{f.inspect}\n" }.join
-        required = ruby_exe(code, options: '--disable-gems')
-        required.should == "false\n" * provided.size
-      end
+      code = provided.map { |f| "puts require #{f.inspect}\n" }.join
+      required = ruby_exe(code, options: '--disable-gems')
+      required.should == "false\n" * provided.size
     end
 
-    ruby_version_is "2.5" do
-      it "complex, enumerator, rational and thread are already required" do
-        provided = %w[complex enumerator rational thread]
-        features = ruby_exe("puts $LOADED_FEATURES", options: '--disable-gems')
-        provided.each { |feature|
-          features.should =~ /\b#{feature}\.(rb|so|jar)$/
-        }
+    it "unicode_normalize is part of core and not $LOADED_FEATURES" do
+      features = ruby_exe("puts $LOADED_FEATURES", options: '--disable-gems')
+      features.lines.each { |feature|
+        feature.should_not include("unicode_normalize")
+      }
 
-        code = provided.map { |f| "puts require #{f.inspect}\n" }.join
-        required = ruby_exe(code, options: '--disable-gems')
-        required.should == "false\n" * provided.size
-      end
-
-      it "unicode_normalize is part of core and not $LOADED_FEATURES" do
-        features = ruby_exe("puts $LOADED_FEATURES", options: '--disable-gems')
-        features.lines.each { |feature|
-          feature.should_not include("unicode_normalize")
-        }
-
-        -> { @object.require("unicode_normalize") }.should raise_error(LoadError)
-      end
+      -> { @object.require("unicode_normalize") }.should raise_error(LoadError)
     end
   end
 
@@ -683,7 +663,7 @@ describe :kernel_require, shared: true do
         Thread.current[:wait_for] = t2
         Thread.current[:con_raise] = true
 
-        lambda {
+        -> {
           @object.require(@path)
         }.should raise_error(RuntimeError)
 
@@ -724,7 +704,7 @@ describe :kernel_require, shared: true do
       t1 = Thread.new do
         Thread.current[:con_raise] = true
 
-        lambda {
+        -> {
           @object.require(@path)
         }.should raise_error(RuntimeError)
 
@@ -764,7 +744,7 @@ describe :kernel_require, shared: true do
   it "stores the missing path in a LoadError object" do
     path = "abcd1234"
 
-    lambda {
+    -> {
       @object.send(@method, path)
     }.should raise_error(LoadError) { |e|
       e.path.should == path
